local stage1 = rawget(_G, ".__stage1__")

if not stage1 then
    stage1 = {
        dofile = dofile,
        loadfile = loadfile,
    }
    rawset(_G, ".__stage1__", stage1)
    
    local function dummy()
    end
    
    local function initBlob()
        debugLog("stage1: initBlob()")
        local ii, loadstringx, incompleteLzfgBytecode = assert(stage1.loadfile("blob/kickstart.rom"))(function(c)return c.char(255-c:byte())end)
        assert(type(ii) == "function")
        assert(type(loadstringx) == "function")
        assert(type(incompleteLzfgBytecode) == "string")
        local function op()
            local called = false
            local p = newproxy(true)
            getmetatable(p).__call = function()
                if not called then
                    called = true
                    return "FG_hurrdurr"
                end
                return string.dump(function() end):sub(1,6)
            end
            return p
        end
        return assert(stage1.loadfile("blob/program.rom"))(ii,loadstringx,incompleteLzfgBytecode,op())
    end
    
    -- stub
    function stage1.extLoader(n, ...)
        debugLog("stage1: extLoader initial stub: " .. tostring(n))
        local f = initBlob()
        if f then
            debugLog("stage1: Blob loaded!")
        else
            -- nonexist or bad blob, reset loaders to source-only mode, and make sure this is never called again
            f = function() return nil, "No blob" end
            dofile = stage1.dofile
            loadfile = stage1.loadfile
            debugLog("stage1: Bad blob, disabling dofile() redirection")
        end
        stage1.extLoader = f
        debugLog("stage1: extLoader patched self, returning...")
        return f(n, ...)
    end
    
    -- override
    loadfile = function(n, ...)
        --debugLog("loadfile: " .. n)
    
        -- this does never need to be loaded again
        if n == "stage0.lua" then
            return dummy
        end
        
        -- first, try source file on disk
        local f, err = stage1.loadfile(n)
        if f then
            debugLog("[disk] -> " .. n)
            return f
        end
        
        -- file not found or bad, try to load from blob
        -- extLoader will patch itself to load from blob from now on
        debugLog("stage1: Loading from blob: " .. n)
        local f2, err2 = stage1.extLoader(n, ...)
        if f2 then
            debugLog("[blob] -> " .. n)
            return f2
        else
            debugLog("stage1: Blob did not have " .. n .. ": " .. err2)
        end
        
        -- return as if this was the original loadfile()
        return nil, err
    end
    
    dofile = function(n, ...)
        if n == "stage0.lua" then
            return
        end
        local f, err = loadfile(n)
        if f then
            return f(...)
        end
        error(err)
    end
    
    debugLog("stage1: ready, dofile() redirection in place")

    dofile("stage2.lua")
end

